# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
import errno
from io import StringIO
import os
from stat import ST_MODE
import tempfile
import unittest
from unittest import mock

from storlets.sbus.client import SBusResponse
from storlets.sbus.client.exceptions import SBusClientIOError
from storlets.gateway.common.exceptions import StorletRuntimeException, \
    StorletTimeout
from storlets.gateway.common.stob import StorletData
from storlets.gateway.gateways.container.gateway import ContainerStorletRequest
from storlets.gateway.gateways.container.runtime import \
    RunTimePaths, StorletInvocationProtocol
from tests.unit import FakeLogger, with_tempdir
from tests.unit.gateway.gateways import FakeFileManager


@contextmanager
def _mock_os_pipe(bufs):
    class FakeFd(object):
        def __init__(self, rbuf=''):
            self.rbuf = rbuf.encode('utf-8')
            self.closed = False

        def read(self, size):
            size = min(len(self.rbuf), size)
            ret = self.rbuf[:size]
            self.rbuf = self.rbuf[size:]
            return ret

        def close(self):
            if self.closed:
                raise OSError(errno.EBADF, os.strerror(errno.EBADF))
            self.closed = True

    def fake_os_read(fd, size):
        return fd.read(size)

    def fake_os_close(fd):
        fd.close()

    pipes = [(FakeFd(buf), FakeFd()) for buf in bufs]
    pipe_generator = iter(pipes)

    def mock_os_pipe():
        try:
            return next(pipe_generator)
        except StopIteration:
            raise AssertionError('pipe called more than expected')

    with mock.patch('storlets.gateway.gateways.container.runtime.os.pipe',
                    mock_os_pipe), \
        mock.patch('storlets.gateway.gateways.container.runtime.os.read',
                   fake_os_read) as fake_os_read,\
        mock.patch('storlets.gateway.gateways.container.runtime.os.close',
                   fake_os_close) as fake_os_close:
        yield pipes


class TestRuntimePaths(unittest.TestCase):

    def setUp(self):
        self.scope = '0123456789abc'
        self._initialize()

    def _initialize(self):
        # TODO(takashi): take these values from config file
        base_dir = '/var/lib/storlets'
        self.script_dir = os.path.join(base_dir, 'scripts')
        self.pipes_dir = os.path.join(base_dir, 'pipes', 'scopes')
        self.storlets_dir = os.path.join(base_dir, 'storlets', 'scopes')
        self.log_dir = os.path.join(base_dir, 'logs', 'scopes')
        self.cache_dir = os.path.join(base_dir, 'cache', 'scopes')

        self.conf = {}
        self.storlet_id = 'org.openstack.storlet.mystorlet'
        self.paths = RunTimePaths(self.scope, self.conf)

    def tearDown(self):
        pass

    def test_host_pipe_dir(self):
        self.assertEqual(
            os.path.join(self.pipes_dir, self.scope),
            self.paths.host_pipe_dir)

    def test_create_host_pipe_dir(self):
        pipedir = self.paths.host_pipe_dir

        # When the directory exists
        with mock.patch('os.path.exists', return_value=True), \
                mock.patch('os.makedirs') as m, \
                mock.patch('os.chmod') as c:
            self.assertEqual(os.path.join(self.pipes_dir, self.scope),
                             self.paths.create_host_pipe_dir())
            self.assertEqual(0, m.call_count)
            cargs, ckwargs = c.call_args
            # Make sure about the target directory
            self.assertEqual(cargs[0], pipedir)

        # When the directory does not exist
        with mock.patch('os.path.exists', return_value=False), \
                mock.patch('os.makedirs') as m, \
                mock.patch('os.chmod') as c:
            self.assertEqual(os.path.join(self.pipes_dir, self.scope),
                             self.paths.create_host_pipe_dir())
            self.assertEqual(1, m.call_count)
            # Make sure about the target directory
            margs, mkwargs = m.call_args
            self.assertEqual(margs[0], pipedir)
            cargs, ckwargs = c.call_args
            self.assertEqual(cargs[0], pipedir)

    def test_host_factory_pipe(self):
        self.assertEqual(
            self.paths.host_factory_pipe,
            os.path.join(self.pipes_dir, self.scope, 'factory_pipe'))

    def test_get_host_storlet_pipe(self):
        self.assertEqual(
            os.path.join(self.pipes_dir, self.scope, self.storlet_id),
            self.paths.get_host_storlet_pipe(self.storlet_id))

    def test_get_sbox_storlet_pipe(self):
        self.assertEqual(
            os.path.join('/mnt/channels', self.storlet_id),
            self.paths.get_sbox_storlet_pipe(self.storlet_id))

    def test_get_sbox_storlet_dir(self):
        self.assertEqual(
            os.path.join('/home/swift', self.storlet_id),
            self.paths.get_sbox_storlet_dir(self.storlet_id))

    def test_host_storlet_base_dir(self):
        self.assertEqual(
            self.paths.host_storlet_base_dir,
            os.path.join(self.storlets_dir, self.scope))

    def test_get_host_storlet_dir(self):
        self.assertEqual(
            os.path.join(self.storlets_dir, self.scope, self.storlet_id),
            self.paths.get_host_storlet_dir(self.storlet_id))

    def test_get_host_slog_path(self):
        self.assertEqual(
            os.path.join(self.log_dir, self.scope, self.storlet_id,
                         'storlet_invoke.log'),
            self.paths.get_host_slog_path(self.storlet_id))

    def test_host_storlet_cache_dir(self):
        self.assertEqual(
            os.path.join(self.cache_dir, self.scope, 'storlet'),
            self.paths.host_storlet_cache_dir)

    def test_host_dependency_cache_dir(self):
        self.assertEqual(
            os.path.join(self.cache_dir, self.scope, 'dependency'),
            self.paths.host_dependency_cache_dir)

    def test_runtime_paths_default(self):
        # CHECK: docs  says we need 4 dirs for communicate
        # ====================================================================
        # |1| host_factory_pipe_path    | <pipes_dir>/<scope>/factory_pipe   |
        # ====================================================================
        # |2| host_storlet_pipe_path    | <pipes_dir>/<scope>/<storlet_id>   |
        # ====================================================================
        # |3| sandbox_factory_pipe_path | /mnt/channels/factory_pipe         |
        # ====================================================================
        # |4| sandbox_storlet_pipe_path | /mnt/channels/<storlet_id>         |
        # ====================================================================
        #
        # With this test,  the scope value is "account" and the storlet_id is
        # "Storlet-1.0.jar" (app name?)
        # ok, let's check for these values

        runtime_paths = RunTimePaths('account', {})
        storlet_id = 'Storlet-1.0.jar'

        # For pipe
        self.assertEqual('/var/lib/storlets/pipes/scopes/account',
                         runtime_paths.host_pipe_dir)

        # 1. host_factory_pipe_path <pipes_dir>/<scope>/factory_pipe
        self.assertEqual(
            '/var/lib/storlets/pipes/scopes/account/factory_pipe',
            runtime_paths.host_factory_pipe)
        # 2. host_storlet_pipe_path <pipes_dir>/<scope>/<storlet_id>
        self.assertEqual(
            '/var/lib/storlets/pipes/scopes/account/Storlet-1.0.jar',
            runtime_paths.get_host_storlet_pipe(storlet_id))
        # 3. Yes, right now, we don't have the path for #3 in Python
        # 4. sandbox_storlet_pipe_path | /mnt/channels/<storlet_id>
        self.assertEqual('/mnt/channels/Storlet-1.0.jar',
                         runtime_paths.get_sbox_storlet_pipe(storlet_id))

        # This looks like for jar load?
        self.assertEqual('/var/lib/storlets/storlets/scopes/account',
                         runtime_paths.host_storlet_base_dir)
        self.assertEqual(
            '/var/lib/storlets/storlets/scopes/account/Storlet-1.0.jar',
            runtime_paths.get_host_storlet_dir(storlet_id))
        # And this one is a mount point in sand box?
        self.assertEqual('/home/swift/Storlet-1.0.jar',
                         runtime_paths.get_sbox_storlet_dir(storlet_id))

    @with_tempdir
    def test_create_host_pipe_dir_with_real_dir(self, temp_dir):
        runtime_paths = RunTimePaths('account', {'host_root': temp_dir})
        runtime_paths.create_host_pipe_dir()
        path = runtime_paths.host_pipe_dir
        self.assertTrue(os.path.exists(path))
        self.assertTrue(os.path.isdir(path))
        permission = oct(os.stat(path)[ST_MODE])[-3:]
        # TODO(kota_): make sure if this is really acceptable
        self.assertEqual('777', permission)


class TestRuntimePathsTempauth(TestRuntimePaths):
    def setUp(self):
        self.scope = 'test'
        self._initialize()


class TestStorletInvocationProtocol(unittest.TestCase):
    def setUp(self):
        self.pipe_path = tempfile.mktemp()
        self.log_file = tempfile.mktemp()
        self.logger = FakeLogger()
        self.storlet_id = 'Storlet-1.0.jar'
        self.options = {'storlet_main': 'org.openstack.storlet.Storlet',
                        'storlet_dependency': 'dep1,dep2',
                        'storlet_language': 'java',
                        'file_manager': FakeFileManager('storlet', 'dep')}
        data = StorletData({}, iter(StringIO()))
        storlet_request = ContainerStorletRequest(
            self.storlet_id, {}, data, options=self.options)
        self.protocol = StorletInvocationProtocol(
            storlet_request, self.pipe_path, self.log_file, 1, self.logger)

    def tearDown(self):
        for path in [self.pipe_path, self.log_file]:
            try:
                os.unlink(path)
            except OSError:
                pass

    def test_send_execute_command(self):
        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.execute') as execute:
            execute.return_value = SBusResponse(True, 'OK', 'someid')
            with self.protocol.storlet_logger.activate():
                self.protocol._send_execute_command()
            self.assertEqual('someid', self.protocol.task_id)

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.execute') as execute:
            execute.return_value = SBusResponse(True, 'OK')
            with self.assertRaises(StorletRuntimeException):
                with self.protocol.storlet_logger.activate():
                    self.protocol._send_execute_command()

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.execute') as execute:
            execute.return_value = SBusResponse(False, 'NG', 'someid')
            with self.assertRaises(StorletRuntimeException):
                with self.protocol.storlet_logger.activate():
                    self.protocol._send_execute_command()

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.execute') as execute:
            execute.side_effect = SBusClientIOError()
            with self.assertRaises(StorletRuntimeException):
                with self.protocol.storlet_logger.activate():
                    self.protocol._send_execute_command()

    def test_invocation_protocol(self):
        # os.pipe will be called 3 times
        pipe_called = 3

        with _mock_os_pipe([''] * pipe_called) as pipes:
            with mock.patch.object(self.protocol,
                                   '_wait_for_read_with_timeout'), \
                    mock.patch.object(self.protocol, '_send_execute_command'):
                self.protocol._invoke()

            self.assertEqual(pipe_called, len(pipes))
            pipes = iter(pipes)

            # data write is not directly closed
            # data read is closed
            input_data_read_fd, input_data_write_fd = next(pipes)
            self.assertTrue(input_data_read_fd.closed)
            self.assertFalse(input_data_write_fd.closed)

            # data write is closed but data read is still open
            data_read_fd, data_write_fd = next(pipes)
            self.assertFalse(data_read_fd.closed)
            self.assertTrue(data_write_fd.closed)

            # metadata write fd is closed, metadata read fd is still open.
            metadata_read_fd, metadata_write_fd = next(pipes)
            self.assertFalse(metadata_read_fd.closed)
            self.assertTrue(metadata_write_fd.closed)

            # sanity
            self.assertRaises(StopIteration, next, pipes)

    def test_invocation_protocol_remote_fds(self):
        # In default, we have 4 fds in remote_fds
        data = StorletData({}, iter(StringIO()))
        storlet_request = ContainerStorletRequest(
            self.storlet_id, {}, data, options=self.options)
        protocol = StorletInvocationProtocol(
            storlet_request, self.pipe_path, self.log_file, 1, self.logger)
        with protocol.storlet_logger.activate():
            self.assertEqual(4, len(protocol.remote_fds))

        # extra_resources expands the remote_fds
        data = StorletData({}, iter(StringIO()))
        extra_data_list = [StorletData({}, iter(StringIO()))]
        storlet_request = ContainerStorletRequest(
            self.storlet_id, {}, data, options=self.options,
            extra_data_list=extra_data_list)
        protocol = StorletInvocationProtocol(
            storlet_request, self.pipe_path, self.log_file, 1, self.logger)
        with protocol.storlet_logger.activate():
            self.assertEqual(5, len(protocol.remote_fds))

        # 2 more extra_resources expands the remote_fds
        data = StorletData({}, iter(StringIO()))
        extra_data_list = [
            StorletData({}, iter(StringIO())),
            StorletData({}, iter(StringIO())),
            StorletData({}, iter(StringIO()))
        ]
        storlet_request = ContainerStorletRequest(
            self.storlet_id, {}, data, options=self.options,
            extra_data_list=extra_data_list)
        protocol = StorletInvocationProtocol(
            storlet_request, self.pipe_path, self.log_file, 1, self.logger)
        with protocol.storlet_logger.activate():
            self.assertEqual(7, len(protocol.remote_fds))

    def test_open_writer_with_invalid_fd(self):
        invalid_fds = (
            (None, TypeError), (-1, ValueError), ('blah', TypeError))

        for invalid_fd, expected_error in invalid_fds:
            with self.assertRaises(expected_error):
                with self.protocol._open_writer(invalid_fd):
                    pass

    def _test_writer_with_exception(self, exception_cls):
        pipes = [os.pipe()]

        def raise_in_the_context():
            with self.protocol._open_writer(pipes[0][1]):
                raise exception_cls()
        try:
            # writer context doesn't suppress any exception
            self.assertRaises(exception_cls, raise_in_the_context)

            # since _open_writer closes the write fd, the os.close will fail as
            # BadFileDescriptor
            with self.assertRaises(OSError) as os_error:
                os.close(pipes[0][1])
            self.assertEqual(9, os_error.exception.errno)

        finally:
            for fd in pipes[0]:
                try:
                    os.close(fd)
                except OSError:
                    pass

    def test_writer_raise_while_in_writer_context(self):
        # basic storlet timeout
        self._test_writer_with_exception(StorletTimeout)
        # unexpected IOError
        self._test_writer_with_exception(IOError)
        # else
        self._test_writer_with_exception(Exception)


class TestStorletInvocationProtocolPython(TestStorletInvocationProtocol):
    def setUp(self):
        self.pipe_path = tempfile.mktemp()
        self.log_file = tempfile.mktemp()
        self.logger = FakeLogger()
        self.storlet_id = 'Storlet-1.0.py'
        self.options = {'storlet_main': 'storlet.Storlet',
                        'storlet_dependency': 'dep1,dep2',
                        'storlet_language': 'python',
                        'language_version': '3.6',
                        'file_manager': FakeFileManager('storlet', 'dep')}
        data = StorletData({}, iter(StringIO()))
        storlet_request = ContainerStorletRequest(
            self.storlet_id, {}, data, options=self.options)
        self.protocol = StorletInvocationProtocol(
            storlet_request, self.pipe_path, self.log_file, 1, self.logger)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swift.common.request_helpers import get_sys_meta_prefix
from swift.common.swob import HTTPMethodNotAllowed, \
    HTTPRequestedRangeNotSatisfiable, Range
from swift.common.utils import public
from storlets.swift_middleware.handlers.base import StorletBaseHandler, \
    NotStorletRequest


class StorletObjectHandler(StorletBaseHandler):
    def __init__(self, request, conf, gateway_conf, app, logger):
        super(StorletObjectHandler, self).__init__(
            request, conf, gateway_conf, app, logger)
        # object need the gateway module only execution
        if (self.is_storlet_execution):
            self._setup_gateway()
        else:
            raise NotStorletRequest()

    def _parse_vaco(self):
        _, _, acc, cont, obj = self.request.split_path(
            5, 5, rest_with_last=True)
        # TODO(kota_): make sure why object server api version is 0?
        return ('0', acc, cont, obj)

    @property
    def is_slo_get_request(self):
        """
        Determines from a GET request and its  associated response
        if the object is a SLO
        """
        return self.request.params.get('multipart-manifest') == 'get'

    def is_symlink_response(self, resp):
        """
        Determines whether the response is a symlink one

        :param resp: swob.Response instance
        :return: Whenther the response is a slo one
        """
        self.logger.debug(
            'Verify if {0} is a symlink'.format(self.path))

        symlink_header = get_sys_meta_prefix('object') + 'symlink-target'
        is_symlink = symlink_header in resp.headers
        if is_symlink:
            self.logger.debug(
                '{0} is indeed a symlink'.format(self.path))
        else:
            self.logger.debug(
                '{0} is NOT a symlink'.format(self.path))
        return is_symlink

    def _get_storlet_invocation_options(self, req):
        options = super(StorletObjectHandler, self).\
            _get_storlet_invocation_options(req)

        # If the request is a storlet request with an single input range, we
        # pass range parameters to storlet gateway, to realize range handling
        # with keepling zero copy
        if self.is_storlet_range_request and \
                not self.is_storlet_multiple_range_request:
            srange = Range(req.headers['X-Storlet-Range'])

            # As we should include the end byte in HTTP Range, here we +1
            # for the end cursor so that we can treat it as general range
            # (include start, and exclude end)
            options['range_start'] = srange.ranges[0][0]
            options['range_end'] = srange.ranges[0][1] + 1

        return options

    def handle_request(self):
        if hasattr(self, self.request.method):
            try:
                handler = getattr(self, self.request.method)
                getattr(handler, 'publicly_accessible')
            except AttributeError:
                # TODO(kota_): add allowed_method list to Allow header
                return HTTPMethodNotAllowed(request=self.request)
            return handler()
        else:
            # un-defined method should be NOT ALLOWED
            raise HTTPMethodNotAllowed(request=self.request)

    def _call_gateway(self, resp):
        sreq = self._build_storlet_request(
            self.request, resp.headers, resp.app_iter)
        return self.gateway.invocation_flow(sreq)

    @public
    def GET(self):
        """
        GET handler on object-server
        """
        # The proxy may add a Range header in the case
        # where the execution is to be done on proxy only
        # (and X-Storlet-Range header exists)
        # Hence we allow having a Range header ONLY
        # if there is also X-Storlet-Range
        # Otherwise, running a Storlet together with
        # The HTTP Range header is not allowed
        if self.is_range_request and not self.is_storlet_range_request:
            raise HTTPRequestedRangeNotSatisfiable(
                b'Storlet execution with range header is not supported',
                request=self.request)

        orig_resp = self.request.get_response(self.app)

        if not orig_resp.is_success:
            return orig_resp

        # TODO(takashi): not sure manifest file should not be run with storlet
        not_runnable = any(
            [self.execute_on_proxy,
             self.execute_range_on_proxy,
             self.is_slo_get_request,
             self.is_slo_response(orig_resp),
             self.is_symlink_response(orig_resp)
             ])

        if not_runnable:
            # Storlet must be invoked on proxy as it is:
            # either an SLO
            # or storlet-range-request
            # or proxy only mode
            self.logger.debug(
                'storlet_handler: invocation over %s to be executed on proxy'
                % self.request.path)
            return orig_resp
        else:
            # We apply here the Storlet:
            self.logger.debug(
                'storlet_handler: invocation over %s to be executed locally'
                % self.request.path)
            return self.apply_storlet(orig_resp)



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
from eventlet import Timeout
from swift.common.swob import HTTPException, HTTPInternalServerError, wsgify
from swift.common.utils import get_logger, register_swift_info
from storlets.gateway.common.exceptions import StorletRuntimeException, \
    StorletTimeout
from storlets.gateway.loader import load_gateway
from storlets.swift_middleware.handlers.base import NotStorletRequest, \
    get_container_names
from storlets.swift_middleware.handlers import StorletProxyHandler, \
    StorletObjectHandler


class StorletHandlerMiddleware(object):

    def __init__(self, app, conf, gateway_conf):
        self.app = app
        self.logger = get_logger(conf, log_route='storlet_handler')
        self.exec_server = conf.get('execution_server')
        self.handler_class = self._get_handler(self.exec_server)
        self.conf = conf
        self.gateway_conf = gateway_conf

    def _get_handler(self, exec_server):
        """
        Generate Handler class based on execution_server parameter

        :param exec_server: Where this storlet_middleware is running.
                            This should value should be 'proxy' or 'object'
        :raise ValueError: If exec_server is invalid
        """
        if exec_server == 'proxy':
            return StorletProxyHandler
        elif exec_server == 'object':
            return StorletObjectHandler
        else:
            raise ValueError(
                'configuration error: execution_server must be either proxy'
                ' or object but is %s' % exec_server)

    @wsgify
    def __call__(self, req):
        try:
            request_handler = self.handler_class(
                req, self.conf, self.gateway_conf, self.app, self.logger)
            self.logger.debug('storlet_handler call in %s: with %s/%s/%s' %
                              (self.exec_server, request_handler.account,
                               request_handler.container, request_handler.obj))
        except HTTPException:
            raise
        except NotStorletRequest:
            return req.get_response(self.app)

        try:
            return request_handler.handle_request()

        # TODO(takashi): Consider handling them in lower layers
        except StorletTimeout:
            self.logger.exception('Storlet execution timed out')
            raise HTTPInternalServerError(body='Storlet execution timed out')
        except StorletRuntimeException:
            self.logger.exception('Storlet execution failed')
            raise HTTPInternalServerError(body='Storlet execution failed')
        except Timeout:
            self.logger.exception('Internal request timed out')
            raise HTTPInternalServerError(body='Internal request timed out')
        except HTTPException:
            # TODO(takashi): Should we generate this log for all error?
            #                (ex. 404 when the object is not found)
            self.logger.exception('Storlet execution failed')
            raise
        except Exception:
            self.logger.exception('Internal server error')
            raise HTTPInternalServerError(body='Internal server error')


def filter_factory(global_conf, **local_conf):
    conf = global_conf.copy()
    conf.update(local_conf)

    module_name = conf.get('storlet_gateway_module', 'stub')
    gateway_class = load_gateway(module_name)
    conf['gateway_module'] = gateway_class

    configParser = configparser.RawConfigParser()
    configParser.read(conf.get('storlet_gateway_conf',
                               '/etc/swift/storlet_stub_gateway.conf'))
    gateway_conf = dict(configParser.items("DEFAULT"))

    # TODO(eranr): Add supported storlets languages and
    #  supported storlet API version
    containers = get_container_names(conf)
    swift_info = {'storlet_container': containers['storlet'],
                  'storlet_dependency': containers['dependency'],
                  'storlet_gateway_class': gateway_class.__name__}
    register_swift_info('storlet_handler', False, **swift_info)

    def storlet_handler_filter(app):
        return StorletHandlerMiddleware(app, conf, gateway_conf)
    return storlet_handler_filter



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swiftclient import client
from swiftclient.exceptions import ClientException
from tests.functional.python import StorletPythonFunctionalTest
import unittest


class TestBrokenStorlet(StorletPythonFunctionalTest):
    def setUp(self, version=None):
        self.storlet_log = 'broken.log'
        self.additional_headers = {}
        super(TestBrokenStorlet, self).setUp(
            storlet_dir='broken',
            storlet_name='broken.py',
            storlet_main='broken.BrokenStorlet',
            storlet_file='source.txt',
            version=version)

    def test_get(self):
        req_headers = {'X-Run-Storlet': self.storlet_name}
        req_headers.update(self.additional_headers)
        with self.assertRaises(ClientException) as cm:
            client.get_object(
                self.url, self.token, self.container, self.storlet_file,
                headers=req_headers)
        e = cm.exception
        self.assertEqual(e.http_status, 503)


class TestBrokenStorletOnProxy(TestBrokenStorlet):
    def setUp(self):
        super(TestBrokenStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import errno
import importlib
import os
import signal
import sys
import uuid
from storlets.sbus import SBus
from storlets.agent.common.server import command_handler, EXIT_FAILURE, \
    CommandSuccess, CommandFailure, SBusServer
from storlets.agent.common.utils import get_logger
from storlets.agent.daemon.files import StorletInputFile, \
    StorletRangeInputFile, StorletOutputFile, StorletLogger


class StorletDaemonLoadError(Exception):
    pass


class StorletDaemon(SBusServer):
    """
    An SBusServer implementation for python storlets applications

    :param storlet_name: the program name string which formatted as
                         'module.class'. Note that nested module
                         (e.g. module.submodule.class) is not allowed.
    :param sbus_path: path string to sbus
    :param logger: a logger instance
    :param pool_size: an integer for concurrency running the storlet apps
    """

    def __init__(self, storlet_name, sbus_path, logger, pool_size):
        super(StorletDaemon, self).__init__(sbus_path, logger)

        self.storlet_name = str(storlet_name)
        try:
            module_name, cls_name = self.storlet_name.split('.')
        except ValueError:
            raise ValueError("Invalid storlet name %s" % storlet_name)

        try:
            module = importlib.import_module(module_name)
            self.storlet_cls = getattr(module, cls_name)
        except (ImportError, AttributeError):
            raise StorletDaemonLoadError(
                "Failed to load storlet %s" % self.storlet_name)

        self.pool_size = pool_size
        self.task_id_to_pid = {}
        self.chunk_size = 16

    def _cleanup_pids(self):
        """
        Remove pids which are already terminated
        """
        terminated = []
        for task_id, daemon_pid in self.task_id_to_pid.items():
            try:
                pid, rc = os.waitpid(daemon_pid, os.WNOHANG)
                if pid or rc:
                    terminated.append(task_id)
            except OSError as err:
                if err.errno == errno.ESRCH:
                    terminated.append(task_id)
                if err.errno == errno.ECHILD:
                    # TODO(takashi): Can we skip checking the remaining ones?
                    terminated.append(task_id)
                else:
                    self.logger.exception('Failed to get the status of '
                                          'the subprocess with pid %d' %
                                          daemon_pid)
        for task_id in terminated:
            self.task_id_to_pid.pop(task_id)

    def _remove_pid(self, pid):
        """
        Remove pid from map dict

        :param pid: the pid of the terminated process
        """
        for task_id, daemon_pid in self.task_id_to_pid.items():
            if daemon_pid == pid:
                self.task_id_to_pid.pop(task_id)
                break

    def _wait_child_process(self):
        """
        Wait until the one of the subprocesses gets terminated
        """
        # We save current length of pid map
        prev_num = len(self.task_id_to_pid)

        # First, we need to remove remaining pids of terminated processes
        self._cleanup_pids()
        if not self.task_id_to_pid or len(self.task_id_to_pid) < prev_num:
            # We don't need have to wait here, when we find
            #  1. we do not have any subprocesses
            #  2. some of the subprocesses are already terminated
            # as the result of cleaning up pid map
            return

        try:
            pid = os.wait()[0]
            self._remove_pid(pid)
        except OSError as e:
            if e.errno == errno.ECHILD:
                # Currently we don't have any subprocesses, so reset the dict
                # here
                self.task_id_to_pid = {}
                pass
            else:
                self.logger.exception('Failed to wait existing subprocesses')

    def _wait_all_child_processes(self):
        self.logger.debug('Wait until all of the subprocesses are '
                          'terminated')
        while len(self.task_id_to_pid):
            self._wait_child_process()

    def _safe_close_files(self, files):
        for fobj in files:
            if not fobj.closed:
                self.logger.warning('Fd %d is not closed inside storlet, '
                                    'so close it' % fobj.fileno())
                fobj.close()

    def _create_input_file(self, st_md, in_md, in_fd):
        start = st_md.get('start')
        end = st_md.get('end')
        if start is not None and end is not None:
            return StorletRangeInputFile(in_md, in_fd, int(start), int(end))
        else:
            return StorletInputFile(in_md, in_fd)

    def _safe_close(self, fileno):
        try:
            os.close(fileno)
        except OSError as e:
            if e.errno != errno.EBADF:
                raise
            pass

    @command_handler
    def execute(self, dtg):
        task_id = str(uuid.uuid4())[:8]
        self.logger.debug('Execute storlet %s (task %s)' %
                          (self.storlet_name, task_id))

        while len(self.task_id_to_pid) >= self.pool_size:
            self._wait_child_process()

        pid = os.fork()
        if pid:
            self.logger.debug('Create a subprocess %d for task %s' %
                              (pid, task_id))
            self.task_id_to_pid[task_id] = pid

            for fd in dtg.invocation_fds:
                # We do not use invocation fds in main process, so close them
                self._safe_close(fd)
        else:
            self._safe_close(dtg.service_out_fd)
            self._execute(dtg)
            sys.exit()

        return CommandSuccess('OK', task_id=task_id)

    def _execute(self, dtg):
        storlet_md = dtg.object_in_storlet_metadata
        params = dtg.params
        in_md = dtg.object_in_metadata
        in_fds = dtg.object_in_fds
        out_md_fds = dtg.object_metadata_out_fds
        out_fds = dtg.object_out_fds
        logger_fd = dtg.logger_out_fd

        self.logger.debug('Started storlet invocation process')
        self.logger.debug(
            'Invocation fds: '
            'in_fds:%s in_md:%s out_md_fds:%s out_fds:%s logger_fd:%s' %
            (in_fds, in_md, out_md_fds, out_fds, logger_fd))

        in_files = [self._create_input_file(st_md, md, in_fd)
                    for st_md, md, in_fd in zip(storlet_md, in_md, in_fds)]

        out_files = [StorletOutputFile(out_md_fd, out_fd)
                     for out_md_fd, out_fd in zip(out_md_fds, out_fds)]

        try:
            self.logger.debug('Start storlet invocation')
            with StorletLogger(self.storlet_name, logger_fd) as slogger:
                handler = self.storlet_cls(slogger)
                handler(in_files, out_files, params)
            self.logger.debug('Completed storlet invocation')
        except Exception:
            self.logger.exception('Error in storlet invocation')
        finally:
            # Make sure that all fds are closed
            self._safe_close_files(in_files + out_files)

    @command_handler
    def cancel(self, dtg):
        task_id = dtg.task_id
        self.logger.debug('Cancel task: %s' % task_id)
        if task_id not in self.task_id_to_pid:
            return CommandFailure('Task id %s is not found' % task_id, False)

        pid = self.task_id_to_pid.get(task_id)
        try:
            os.kill(pid, signal.SIGKILL)
            self._remove_pid(pid)
            return CommandSuccess('Cancelled task %s' % task_id, False)
        except OSError:
            self.logger.exception('Failed to kill subprocess: %d' % pid)
            return CommandFailure('Failed to cancel task %s' % task_id, False)

    @command_handler
    def halt(self, dtg):
        self.logger.debug('Terminating')
        return CommandSuccess('OK', False)

    def _terminate(self):
        self._wait_all_child_processes()

    def _force_terminate(self):
        os.killpg(0, signal.SIGKILL)


def main():
    """
    The entry point of daemon_factory process
    """
    parser = argparse.ArgumentParser(
        description='Daemon process to execute storlet applications')
    parser.add_argument('storlet_name', help='storlet name')
    parser.add_argument('sbus_path', help='the path to unix domain socket')
    parser.add_argument('log_level', help='log level')
    parser.add_argument('pool_size', type=int,
                        help='the maximum thread numbers used swapns for '
                             'one storlet application')
    parser.add_argument('container_id', help='container id')
    opts = parser.parse_args()

    # Initialize logger
    logger = get_logger("storlets-daemon", opts.log_level, opts.container_id)
    logger.debug("Storlet Daemon started")

    try:
        SBus.start_logger(opts.log_level.upper(),
                          container_id=opts.container_id)

        # create an instance of storlet daemon
        daemon = StorletDaemon(opts.storlet_name, opts.sbus_path,
                               logger, opts.pool_size)

        # Start the main loop
        sys.exit(daemon.main_loop())

    except Exception:
        logger.exception('Unhandled exception')
        sys.exit(EXIT_FAILURE)



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from storlets.tools.testtools import FakeStorletFileIn, FakeStorletFileOut


class TestTestTools(TestCase):
    def test_fake_storlet_file_in(self):
        input_string = 'abcdefghijklmonp'
        input_metadata = {'foo': 'bar'}
        fake_storlet_in = FakeStorletFileIn(input_string, input_metadata)
        self.assertEqual(input_string, fake_storlet_in.read())
        self.assertEqual(input_metadata, fake_storlet_in.get_metadata())

        fake_storlet_in.close()
        self.assertTrue(fake_storlet_in.closed)

    def test_fake_storlet_file_out(self):
        out_string = 'abcdefghijklmonp'
        fake_storlet_out = FakeStorletFileOut()
        fake_storlet_out.write(out_string)
        self.assertEqual(out_string, fake_storlet_out.read())
        fake_storlet_out.write(out_string)
        self.assertEqual(out_string * 2, fake_storlet_out.read())

        self.assertIsNone(fake_storlet_out._metadata)
        fake_storlet_out.set_metadata({'test': 'tester'})
        self.assertEqual({'test': 'tester'}, fake_storlet_out._metadata)

        fake_storlet_out.close()
        self.assertTrue(fake_storlet_out.closed)

    def test_fake_storlet_file_out_set_metadata_twice_will_cause_error(self):
        fake_storlet_out = FakeStorletFileOut()
        # first call, it's ok
        fake_storlet_out.set_metadata({})
        # second call will raise an exception
        with self.assertRaises(IOError) as cm:
            fake_storlet_out.set_metadata({})
        self.assertEqual(
            "Sending metadata twice is not allowed",
            cm.exception.args[0])



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ctypes import c_char_p, c_int, c_float, CDLL, POINTER
from storlets.sbus.datagram import build_datagram_from_raw_message


class SBus(object):
    """
    Wrapper class for low level C-API for SBus functionality

    """
    SBUS_SO_NAME = '/usr/local/lib/storlets/libsbus.so'

    def __init__(self):
        # load the C-library
        self.sbus_back_ = CDLL(SBus.SBUS_SO_NAME)

        # create SBus
        self.sbus_back_.sbus_create.argtypes = [c_char_p]
        self.sbus_back_.sbus_create.restype = c_int

        # listen to SBus
        self.sbus_back_.sbus_listen.argtypes = [c_int, c_float]
        self.sbus_back_.sbus_listen.restype = c_int

        # send message
        self.sbus_back_.sbus_send_msg.argtypes = [c_char_p,
                                                  POINTER(c_int),
                                                  c_int,
                                                  c_char_p,
                                                  c_int,
                                                  c_char_p,
                                                  c_int]
        self.sbus_back_.sbus_send_msg.restype = c_int

        # receive message
        self.sbus_back_.sbus_recv_msg.argtypes = [c_int,
                                                  POINTER(POINTER(c_int)),
                                                  POINTER(c_int),
                                                  POINTER(c_char_p),
                                                  POINTER(c_int),
                                                  POINTER(c_char_p),
                                                  POINTER(c_int)]
        self.sbus_back_.sbus_recv_msg.restype = c_int

    @staticmethod
    def start_logger(str_log_level='DEBUG', container_id=None):
        # load the C-library
        sbus_back_ = CDLL(SBus.SBUS_SO_NAME)

        sbus_back_.sbus_start_logger.argtypes = [c_char_p, c_char_p]
        sbus_back_.sbus_start_logger(str_log_level.encode("utf-8"),
                                     container_id.encode("utf-8"))

    @staticmethod
    def stop_logger():
        # load the C-library
        sbus_back_ = CDLL(SBus.SBUS_SO_NAME)
        sbus_back_.sbus_stop_logger()

    def create(self, sbus_name):
        return self.sbus_back_.sbus_create(sbus_name.encode("utf-8"))

    def listen(self, sbus_handler, timeout=0.0):
        return self.sbus_back_.sbus_listen(sbus_handler, timeout)

    def receive(self, sbus_handler):
        ph_files = POINTER(c_int)()
        pp_metadata = (c_char_p)()
        pp_params = (c_char_p)()
        pn_files = (c_int)()
        pn_metadata = (c_int)()
        pn_params = (c_int)()

        # Invoke C function
        n_status = self.sbus_back_.sbus_recv_msg(sbus_handler,
                                                 ph_files,
                                                 pn_files,
                                                 pp_metadata,
                                                 pn_metadata,
                                                 pp_params,
                                                 pn_params)
        if n_status < 0:
            return None

        # The invocation was successful.
        # De-serialize the data

        # Aggregate file descriptors
        n_files = pn_files.value
        h_files = []
        for i in range(n_files):
            h_files.append(ph_files[i])

        # Extract Python strings
        n_metadata = pn_metadata.value
        str_metadata = pp_metadata.value.decode("utf-8")
        n_params = pn_params.value
        str_params = pp_params.value.decode("utf-8")

        # Trim the junk out
        if 0 < n_metadata:
            str_metadata = str_metadata[0:n_metadata]
        str_params = str_params[0:n_params]

        # Construct actual result datagram
        return build_datagram_from_raw_message(
            h_files, str_metadata, str_params)

    @staticmethod
    def send(sbus_name, datagram):
        # Serialize the datagram into JSON strings and C integer array
        str_json_params = datagram.serialized_cmd_params
        p_params = c_char_p(str_json_params.encode("utf-8"))
        n_params = c_int(len(str_json_params))

        n_files = c_int(0)
        h_files = None
        n_metadata = c_int(0)
        p_metadata = None

        if datagram.num_fds > 0:
            str_json_metadata = datagram.serialized_metadata
            p_metadata = c_char_p(str_json_metadata.encode("utf-8"))
            n_metadata = c_int(len(str_json_metadata))

            n_fds = datagram.num_fds
            n_files = c_int(n_fds)

            file_fds = datagram.fds
            h_files = (c_int * n_fds)()

            for i in range(n_fds):
                h_files[i] = file_fds[i]

        # Invoke C function
        sbus = SBus()
        n_status = sbus.sbus_back_.sbus_send_msg(
            sbus_name.encode("utf-8"),
            h_files,
            n_files,
            p_metadata,
            n_metadata,
            p_params,
            n_params)
        return n_status



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import uuid

from swiftclient import client as swiftclient
from storlets.tools.cluster_config_parser import ClusterConfig
from storlets.tools.utils import deploy_storlet, get_admin_auth, put_local_file
import os

CONFIG_DIR = os.environ.get('CLUSTER_CONF_DIR', os.getcwd())
CONFIG_FILE = os.path.join(CONFIG_DIR, 'test.conf')
PATH_TO_STORLETS = os.environ.get(
    'STORLET_SAMPLE_PATH',
    # assuming, current working dir is at top of storlet repo
    os.path.join(os.getcwd(), 'StorletSamples'))
CONSOLE_TIMEOUT = 2


class StorletBaseFunctionalTest(unittest.TestCase):
    def setUp(self):
        self.conf_file = CONFIG_FILE
        try:
            self.conf = ClusterConfig(CONFIG_FILE)
        except IOError:
            self.fail('cluster_config.json not found in %s. '
                      'Please check your testing environment.' % CONFIG_DIR)

        self.url, self.token = get_admin_auth(self.conf)
        # TODO(kota_): do we need to call setUp() when inheriting TestCase
        # directly? AFAIK, no setUp method in the class...
        super(StorletBaseFunctionalTest, self).setUp()


class StorletFunctionalTest(StorletBaseFunctionalTest):

    def create_container(self, container):
        response = dict()
        swiftclient.put_container(self.url, self.token,
                                  container, headers=None,
                                  response_dict=response)
        status = response.get('status')
        assert (status >= 200 or status < 300)

    def cleanup_container(self, container):
        # list all objects inside the container
        _, objects = swiftclient.get_container(
            self.url, self.token, container, full_listing=True)

        # delete all objects inside the container
        # N.B. this cleanup could run in parallel but currently we have a few
        # objects in the user testing container so that, currently this does
        # as sequential simply
        for obj_dict in objects:
            swiftclient.delete_object(
                self.url, self.token, container, obj_dict['name'])
        swiftclient.get_container(self.url, self.token, container)

        # delete the container
        swiftclient.delete_container(self.url, self.token, container)

    def setUp(self, language, path_to_bundle,
              storlet_dir,
              storlet_name, storlet_main, storlet_file,
              dep_names, headers, version=None):
        super(StorletFunctionalTest, self).setUp()
        self.storlet_dir = storlet_dir
        self.storlet_name = storlet_name
        self.storlet_main = storlet_main
        self.dep_names = dep_names
        self.path_to_bundle = path_to_bundle
        self.container = 'container-%s' % uuid.uuid4()
        self.storlet_file = storlet_file
        self.headers = headers or {}
        self.acct = self.url.split('/')[4]
        self.deps = []
        if dep_names:
            for d in self.dep_names:
                self.deps.append('%s/%s' % (self.path_to_bundle, d))
        storlet = '%s/%s' % (self.path_to_bundle, self.storlet_name)

        deploy_storlet(self.url, self.token,
                       storlet, self.storlet_main,
                       self.deps, language, version)

        self.create_container(self.container)
        if self.storlet_file:
            put_local_file(self.url, self.token,
                           self.container,
                           self.path_to_bundle,
                           self.storlet_file,
                           self.headers)

    def tearDown(self):
        self.cleanup_container(self.container)



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from stevedore import driver
from storlets.gateway.common.exceptions import StorletGatewayLoadError


def load_gateway(gateway_name):
    """
    Load gateway class for the given name

    :param gateway_name: a name of the gateway class to be loaded
                         you can specify its entry point name or full path
    :returns: a gateway class loaded
    :raises StorletGatewayLoadError: when it fails to load the given gateway
    """

    namespace = 'storlets.gateways'
    try:
        # Try to load gateway class using entry point
        driver_manager = driver.DriverManager(namespace, gateway_name,
                                              invoke_on_load=False)
        return driver_manager.driver
    except RuntimeError:
        # Currently we can not find the entry point, so try again to load
        # using full class path
        mod_str, _sep, class_str = gateway_name.rpartition('.')
        if not mod_str:
            raise StorletGatewayLoadError(
                'Invalid class path is given for gateway class name %s' %
                gateway_name)
        try:
            __import__(mod_str)
            return getattr(sys.modules[mod_str], class_str)
        except (ImportError, AttributeError):
            raise StorletGatewayLoadError('Failed to load gateway %s' %
                                          gateway_name)



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class FileManager(object):
    """
    This class is used to load/save files required for storlet execution
    from/into the storage which also stores data to be processed
    """

    def __init__(self):
        pass

    def get_storlet(self, name):
        """
        Load storlet file content

        :param name: storlet file name
        """
        raise NotImplementedError()

    def get_dependency(self, name):
        """
        Load dependency file content

        :param name: dependency file name
        """
        raise NotImplementedError()

    def put_log(self, name, fobj):
        """
        Save storlet log file to storage

        :param name: log file name
        :param data_iter: File Object
        """
        raise NotImplementedError()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import unittest
from unittest import mock

from contextlib import contextmanager
from swift.common.swob import Request, HTTPOk, HTTPCreated, HTTPAccepted, \
    HTTPNoContent, HTTPNotFound
from storlets.swift_middleware.handlers import StorletProxyHandler
from storlets.swift_middleware.handlers.proxy import REFERER_PREFIX

from tests.unit.swift_middleware.handlers import \
    BaseTestStorletMiddleware, create_handler_config


@contextmanager
def fake_acc_info(acc_info):
    with mock.patch('storlets.swift_middleware.handlers.proxy.'
                    'get_account_info') as ai:
        ai.return_value = acc_info
        yield


@contextmanager
def storlet_enabled():
    acc_info = {'meta': {'storlet-enabled': 'true'}}
    with fake_acc_info(acc_info):
        yield


class TestStorletMiddlewareProxy(BaseTestStorletMiddleware):
    def setUp(self):
        super(TestStorletMiddlewareProxy, self).setUp(exec_server='proxy')

    def test_load_app(self):
        try:
            self.get_app(self.base_app, self.conf)
        except Exception:
            self.fail('Application loading got an error')

    def get_request_response(self, target, method, headers=None, body=None):
        # Ensure the body is byte format on py3, this is needed until
        # swift's Request supports byte format body when body is None in args
        body = body or b''
        req = Request.blank(target, environ={'REQUEST_METHOD': method},
                            headers=headers, body=body)
        return self.get_response(req)

    def test_GET_without_storlets(self):
        def basic_get(path):
            self.base_app.register('GET', path, HTTPOk, body=b'FAKE APP')
            resp = self.get_request_response(path, 'GET')
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE APP', resp.body)
            self.base_app.reset_all()

        for target in ('AUTH_a', 'AUTH_a/c', 'AUTH_a/c/o'):
            path = '/'.join(['', 'v1', target])
            basic_get(path)

    def test_GET_with_storlets(self):
        # TODO(takashi): decide request path based on config value
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk, body=b'FAKE RESULT')
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, headers={},
                               body=b'jar binary')

        acc_info = {'meta': {'storlet-enabled': 'true'}}
        with fake_acc_info(acc_info):
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE RESULT', resp.body)
            calls = self.base_app.get_calls()

            # Make sure now we sent two requests to swift
            self.assertEqual(2, len(calls))

            # The first one is HEAD request to storlet object
            self.assertEqual('HEAD', calls[0][0])
            self.assertEqual(storlet, calls[0][1])

            # The last one is execution GET call
            self.assertEqual(target, calls[-1][1])
            self.assertIn('X-Run-Storlet', calls[-1][2])

    def test_GET_with_storlets_disabled_account(self):
        target = '/v1/AUTH_a/c/o'

        acc_info = {'meta': {}}
        with fake_acc_info(acc_info):
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('400 Bad Request', resp.status)

            calls = self.base_app.get_calls()
            self.assertEqual(0, len(calls))

    def test_GET_with_storlets_object_404(self):
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPNotFound)
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('404 Not Found', resp.status)

            calls = self.base_app.get_calls()
            self.assertEqual(2, len(calls))

    def test_GET_with_storlets_and_http_range(self):
        target = '/v1/AUTH_a/c/o'

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                       'Range': 'bytes=10-20'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('400 Bad Request', resp.status)

    def test_GET_with_storlets_and_storlet_range(self):
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk, body=b'FAKE APP')
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            req_range = 'bytes=1-6'
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Run-On-Proxy': '',
                       'X-Storlet-Range': req_range}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'AKE AP', resp.body)

            self.assertNotIn('Content-Range', resp.headers)
            self.assertEqual('bytes 1-6/8',
                             resp.headers['Storlet-Input-Range'])

            raw_req = self.base_app.get_calls('GET', target)[0]
            for key in ['Range', 'X-Storlet-Range']:
                self.assertEqual(raw_req[2][key], req_range)

    def test_GET_with_storlets_and_object_storlet_range(self):
        # Create a single range request that needs to be
        # processed by the object handler
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk, body=b'FAKE APP')
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            req_range = 'bytes=1-6'
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Range': req_range}
            resp = self.get_request_response(target, 'GET', headers=headers)
            # We assert that nothing actually happens
            # by the proxy handler
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE APP', resp.body)

    def test_GET_with_storlets_and_extra_resource(self):
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk, body=b'FAKE APP')
        extra_target = '/v1/AUTH_a/c2/o2'
        self.base_app.register('GET', extra_target, HTTPOk, body=b'Whooa')
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Extra-Resources': '/c2/o2'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE APP', resp.body)

            # GET target called
            self.assertTrue(any(self.base_app.get_calls('GET', target)))
            # GET extra target also called
            self.assertTrue(any(self.base_app.get_calls('GET', extra_target)))

    def test_PUT_with_storlets_and_extra_resource(self):
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('PUT', target, HTTPCreated, body=b'')
        extra_target = '/v1/AUTH_a/c2/o2'
        self.base_app.register('GET', extra_target, HTTPOk, body=b'Whooa')
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Extra-Resources': '/c2/o2'}
            resp = self.get_request_response(target, 'PUT', headers=headers)
            self.assertEqual('201 Created', resp.status)

            # PUT target called
            self.assertTrue(any(self.base_app.get_calls('PUT', target)))
            # GET extra target also called
            self.assertTrue(any(self.base_app.get_calls('GET', extra_target)))

    def test_GET_with_storlets_and_extra_resource_within_limit(self):
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk, body=b'FAKE APP')
        extra_targets = ['/v1/AUTH_a/c2/o2', '/v1/AUTH_a/c3/o3']
        for extra_target in extra_targets:
            self.base_app.register('GET', extra_target, HTTPOk, body=b'foo')
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        self.conf['max_extra_resources'] = 2

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Extra-Resources': '/c2/o2,/c3/o3'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE APP', resp.body)

            # GET target called
            self.assertTrue(any(self.base_app.get_calls('GET', target)))
            # GET extra targets also called
            for extra_target in extra_targets:
                self.assertTrue(
                    any(self.base_app.get_calls('GET', extra_target)))

    def test_GET_with_storlets_and_extra_resource_not_object(self):
        target = '/v1/AUTH_a/c/o'

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Extra-Resources': '/c2o2'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('400 Bad Request', resp.status)

    def test_GET_with_storlets_and_extra_resource_over_limit(self):
        target = '/v1/AUTH_a/c/o'
        self.conf['max_extra_resources'] = 1

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Extra-Resources': '/c2/o2,/c3/o3'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('400 Bad Request', resp.status)

    def test_GET_with_storlets_and_extra_resource_disabled(self):
        target = '/v1/AUTH_a/c/o'
        self.conf['max_extra_resources'] = 0

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Extra-Resources': '/c2/o2'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('403 Forbidden', resp.status)

    def test_GET_slo_without_storlets(self):
        target = '/v1/AUTH_a/c/slo_manifest'
        self.base_app.register('GET', target, HTTPOk,
                               headers={'x-static-large-object': 'True'},
                               body=b'FAKE APP')
        resp = self.get_request_response(target, 'GET')
        self.assertEqual(b'FAKE APP', resp.body)

    def test_GET_slo_with_storlets(self):
        target = '/v1/AUTH_a/c/slo_manifest'
        self.base_app.register('GET', target, HTTPOk,
                               headers={'x-static-large-object': 'True'},
                               body=b'FAKE APP')
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE APP', resp.body)

            calls = self.base_app.get_calls()
            self.assertEqual(2, len(calls))

    def test_GET_symlink_with_storlets(self):
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk,
                               headers={'x-symlink-target': 'c/o2'},
                               body=b'FAKE APP')
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE APP', resp.body)

            calls = self.base_app.get_calls()
            self.assertEqual(2, len(calls))

    def test_GET_with_storlets_no_object(self):
        target = '/v1/AUTH_a/c/'
        self.base_app.register('GET', target, HTTPOk,
                               body=b'FAKE APP')
        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar'}
            resp = self.get_request_response(target, 'GET', headers=headers)
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE APP', resp.body)

            # Make sure now we sent one request to swift
            # that is, storlets path was not executed
            # due to no object in target
            calls = self.base_app.get_calls()
            self.assertEqual(1, len(calls))

    def test_PUT_without_storlets(self):
        def basic_put(path):
            self.base_app.register('PUT', path, HTTPCreated)
            resp = self.get_request_response(path, 'PUT')
            self.assertEqual('201 Created', resp.status)
            self.base_app.reset_all()

        for target in ('AUTH_a', 'AUTH_a/c', 'AUTH_a/c/o'):
            path = '/'.join(['', 'v1', target])
            basic_put(path)

    def test_PUT_with_storlets(self):
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('PUT', target, HTTPCreated, body=b'')
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar'}
            resp = self.get_request_response(target, 'PUT', headers=headers,
                                             body=b'FAKE APP')
            self.assertEqual('201 Created', resp.status)

            calls = self.base_app.get_calls()
            # Make sure now we sent two requests to swift
            self.assertEqual(2, len(calls))

            # The first one is HEAD request to storlet object
            self.assertEqual('HEAD', calls[0][0])
            self.assertEqual(storlet, calls[0][1])

            # The last one is PUT request about processed object
            self.assertEqual('PUT', calls[-1][0])
            self.assertEqual(target, calls[-1][1])
            self.assertEqual(b'FAKE APP', calls[-1][3])

    def test_PUT_with_storlets_no_object(self):
        target = '/v1/AUTH_a/c/'
        self.base_app.register('PUT', target, HTTPCreated)
        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar'}
            resp = self.get_request_response(target, 'PUT', headers=headers,
                                             body=b'FAKE APP')
            self.assertEqual('201 Created', resp.status)

            calls = self.base_app.get_calls()
            # Make sure now we sent one request to swift
            # that is, storlets path was not executed
            # due to no object in target
            self.assertEqual(1, len(calls))

    def test_PUT_copy_without_storlets(self):
        target = '/v1/AUTH_a/c/to'
        copy_from = 'c/so'
        self.base_app.register('PUT', target, HTTPCreated)

        headers = {'X-Copy-From': copy_from,
                   'X-Backend-Storage-Policy-Index': 0}
        resp = self.get_request_response(target, 'PUT', headers=headers)
        self.assertEqual('201 Created', resp.status)

    def test_PUT_copy_with_storlets(self):
        source = '/v1/AUTH_a/c/so'
        target = '/v1/AUTH_a/c/to'
        copy_from = 'c/so'
        self.base_app.register('GET', source, HTTPOk,
                               headers={'x-object-meta-name': 'name'},
                               body=b'source body')
        self.base_app.register('PUT', target, HTTPCreated)
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'X-Copy-From': copy_from,
                       'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Backend-Storage-Policy-Index': 0}
            resp = self.get_request_response(target, 'PUT', headers=headers)
            self.assertEqual('201 Created', resp.status)
            get_calls = self.base_app.get_calls('GET', source)
            self.assertEqual(1, len(get_calls))
            self.assertEqual(b'', get_calls[-1][3])
            self.assertEqual(source, get_calls[-1][1])
            self.assertIn('X-Run-Storlet', get_calls[-1][2])
            put_calls = self.base_app.get_calls('PUT', target)
            self.assertEqual(1, len(put_calls))
            self.assertEqual(b'source body', put_calls[-1][3])
            self.assertIn('X-Object-Meta-Name', dict(put_calls[-1][2]))
            self.assertEqual('name', put_calls[-1][2]['X-Object-Meta-Name'])
            self.assertNotIn('X-Run-Storlet', put_calls[-1][2])
            # no invocation (at gateway stub) at proxy
            for debug_line in self.logger.get_log_lines('debug'):
                self.assertNotIn("Identity invocation is called", debug_line)

    def test_PUT_copy_with_storlets_run_on_proxy(self):
        source = '/v1/AUTH_a/c/so'
        target = '/v1/AUTH_a/c/to'
        copy_from = 'c/so'
        self.base_app.register('GET', source, HTTPOk,
                               headers={'x-object-meta-name': 'name'},
                               body=b'source body')
        self.base_app.register('PUT', target, HTTPCreated)
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'X-Copy-From': copy_from,
                       'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Run-On-Proxy': '',
                       'X-Backend-Storage-Policy-Index': 0}
            resp = self.get_request_response(target, 'PUT', headers=headers)
            self.assertEqual('201 Created', resp.status)
            get_calls = self.base_app.get_calls('GET', source)
            self.assertEqual(1, len(get_calls))
            self.assertEqual(b'', get_calls[-1][3])
            self.assertEqual(source, get_calls[-1][1])
            self.assertNotIn('X-Run-Storlet', get_calls[-1][2])
            put_calls = self.base_app.get_calls('PUT', target)
            self.assertEqual(1, len(put_calls))
            self.assertEqual(b'source body', put_calls[-1][3])
            self.assertIn('X-Object-Meta-Name', dict(put_calls[-1][2]))
            self.assertEqual('name', put_calls[-1][2]['X-Object-Meta-Name'])
            self.assertNotIn('X-Run-Storlet', put_calls[-1][2])
            # no invocation at proxy
            for debug_line in self.logger.get_log_lines('debug'):
                if "Identity invocation is called" in debug_line:
                    break
            else:
                self.fail('no invocation message found at proxy')

    def test_COPY_verb_without_storlets(self):
        source = '/v1/AUTH_a/c/so'
        destination = 'c/to'
        self.base_app.register('COPY', source, HTTPCreated)

        headers = {'Destination': destination,
                   'X-Backend-Storage-Policy-Index': 0}
        resp = self.get_request_response(source, 'COPY', headers=headers)
        self.assertEqual('201 Created', resp.status)

    def test_COPY_verb_with_storlets(self):
        source = '/v1/AUTH_a/c/so'
        target = '/v1/AUTH_a/c/to'
        destination = 'c/to'
        self.base_app.register('GET', source, HTTPOk,
                               headers={'x-object-meta-name': 'name'},
                               body=b'source body')
        self.base_app.register('PUT', target, HTTPCreated)
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'Destination': destination,
                       'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Backend-Storage-Policy-Index': 0}
            resp = self.get_request_response(source, 'COPY', headers=headers)
            self.assertEqual('201 Created', resp.status)
            get_calls = self.base_app.get_calls('GET', source)
            self.assertEqual(1, len(get_calls))
            self.assertEqual(b'', get_calls[-1][3])
            self.assertEqual(source, get_calls[-1][1])
            self.assertIn('X-Run-Storlet', get_calls[-1][2])
            put_calls = self.base_app.get_calls('PUT', target)
            self.assertEqual(1, len(put_calls))
            self.assertEqual(b'source body', put_calls[-1][3])
            self.assertEqual('name', put_calls[-1][2]['X-Object-Meta-Name'])
            self.assertNotIn('X-Run-Storlet', put_calls[-1][2])
            # no invocation at proxy
            for debug_line in self.logger.get_log_lines('debug'):
                self.assertNotIn("Identity invocation is called", debug_line)

    def test_COPY_verb_with_storlets_run_on_porxy(self):
        source = '/v1/AUTH_a/c/so'
        target = '/v1/AUTH_a/c/to'
        destination = 'c/to'
        self.base_app.register('GET', source, HTTPOk,
                               headers={'x-object-meta-name': 'name'},
                               body=b'source body')
        self.base_app.register('PUT', target, HTTPCreated)
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        with storlet_enabled():
            headers = {'Destination': destination,
                       'X-Run-Storlet': 'Storlet-1.0.jar',
                       'X-Storlet-Run-On-Proxy': '',
                       'X-Backend-Storage-Policy-Index': 0}
            resp = self.get_request_response(source, 'COPY', headers=headers)
            self.assertEqual('201 Created', resp.status)
            get_calls = self.base_app.get_calls('GET', source)
            self.assertEqual(1, len(get_calls))
            self.assertEqual(b'', get_calls[-1][3])
            self.assertEqual(source, get_calls[-1][1])
            self.assertNotIn('X-Run-Storlet', get_calls[-1][2])
            put_calls = self.base_app.get_calls('PUT', target)
            self.assertEqual(1, len(put_calls))
            self.assertEqual(b'source body', put_calls[-1][3])
            self.assertEqual('name', put_calls[-1][2]['X-Object-Meta-Name'])
            self.assertNotIn('X-Run-Storlet', put_calls[-1][2])
            # no invocation at proxy
            for debug_line in self.logger.get_log_lines('debug'):
                if "Identity invocation is called" in debug_line:
                    break
            else:
                self.fail('no invocation message found at proxy')

    def test_copy_with_unsupported_headers(self):
        target = '/v1/AUTH_a/c/o'
        storlet = '/v1/AUTH_a/storlet/Storlet-1.0.jar'
        self.base_app.register('GET', storlet, HTTPOk, body=b'jar binary')

        def copy_400(method, copy_header):
            base_headers = {'X-Run-Storlet': 'Storlet-1.0.jar',
                            'X-Backend-Storage-Policy-Index': 0}
            base_headers.update(copy_header)
            resp = self.get_request_response(target, method,
                                             headers=base_headers)
            self.assertEqual('400 Bad Request', resp.status)

        cases = [('COPY', {'Destination-Account': 'a', 'Destination': 'c/o'}),
                 ('COPY', {'X-Fresh-Metadata': '', 'Destination': 'c/o'}),
                 ('PUT', {'X-Copy-From-Account': 'a', 'X-Copy-From': 'c/o'}),
                 ('PUT', {'X-Fresh-Metadata': '', 'X-Copy-From': 'c/o'})]

        for case in cases:
            with storlet_enabled():
                copy_400(case[0], case[1])

    def test_PUT_storlet(self):
        target = '/v1/AUTH_a/storlet/storlet-1.0.jar'
        self.base_app.register('PUT', target, HTTPCreated)

        with storlet_enabled():
            sheaders = {'X-Object-Meta-Storlet-Language': 'Java',
                        'X-Object-Meta-Storlet-Interface-Version': '1.0',
                        'X-Object-Meta-Storlet-Dependency': 'dependency',
                        'X-Object-Meta-Storlet-Main':
                            'org.openstack.storlet.Storlet'}
            resp = self.get_request_response(target, 'PUT', headers=sheaders)
            self.assertEqual('201 Created', resp.status)

    def test_PUT_storlet_mandatory_parameter_fails(self):
        target = '/v1/AUTH_a/storlet/storlet-1.0.jar'
        self.base_app.register('PUT', target, HTTPCreated)

        drop_headers = [
            ('X-Object-Meta-Storlet-Language', 'Language'),
            ('X-Object-Meta-Storlet-Interface-Version', 'Interface-Version'),
            ('X-Object-Meta-Storlet-Dependency', 'Dependency'),
            ('X-Object-Meta-Storlet-Main', 'Main'),
        ]

        def put(path, header, assertion):
            sheaders = {'X-Object-Meta-Storlet-Language': 'Java',
                        'X-Object-Meta-Storlet-Interface-Version': '1.0',
                        'X-Object-Meta-Storlet-Dependency': 'dependency',
                        'X-Object-Meta-Storlet-Main':
                            'org.openstack.storlet.Storlet'}
            del sheaders[header]
            resp = self.get_request_response(path, 'PUT', headers=sheaders)
            # FIXME(kota_): Unfortunately, we can not test yet here because
            # the validation is not in stub gateway but in docker gateway so
            # need more refactor to parse the functionality to be easy testing
            # self.assertEqual('400 BadRequest', resp.status)
            self.assertEqual('201 Created', resp.status)

        for header, assertion in drop_headers:
            with storlet_enabled():
                put(target, header, assertion)

    def test_POST_storlet(self):
        target = '/v1/AUTH_a/storlet/storlet-1.0.jar'
        self.base_app.register('POST', target, HTTPAccepted)

        with storlet_enabled():
            sheaders = {'X-Object-Meta-Storlet-Language': 'Java',
                        'X-Object-Meta-Storlet-Interface-Version': '1.0',
                        'X-Object-Meta-Storlet-Dependency': 'dependency',
                        'X-Object-Meta-Storlet-Main':
                            'org.openstack.storlet.Storlet'}
            resp = self.get_request_response(target, 'POST', headers=sheaders)
            self.assertEqual('202 Accepted', resp.status)

    def test_GET_storlet(self):
        target = '/v1/AUTH_a/storlet/storlet-1.0.jar'
        sheaders = {'X-Object-Meta-Storlet-Language': 'Java',
                    'X-Object-Meta-Storlet-Interface-Version': '1.0',
                    'X-Object-Meta-Storlet-Dependency': 'dependency',
                    'X-Object-Meta-Storlet-Main':
                        'org.openstack.storlet.Storlet'}
        self.base_app.register('GET', target, HTTPOk, headers=sheaders,
                               body=b'jar binary')

        with storlet_enabled():
            resp = self.get_request_response(target, 'GET')
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'jar binary', resp.body)
            for key in sheaders:
                self.assertEqual(sheaders[key], resp.headers.get(key))

    def test_PUT_dependency(self):
        target = '/v1/AUTH_a/dependency/dependency'
        self.base_app.register('PUT', target, HTTPCreated)

        with storlet_enabled():
            sheaders = {'X-Object-Meta-Storlet-Dependency-Version': '1',
                        'X-Object-Meta-Storlet-Dependency-Permissions': '0755'}
            resp = self.get_request_response(target, 'PUT', headers=sheaders)
            self.assertEqual('201 Created', resp.status)

    def test_POST_dependency(self):
        target = '/v1/AUTH_a/dependency/dependency'
        self.base_app.register('POST', target, HTTPAccepted)

        with storlet_enabled():
            sheaders = {'X-Object-Meta-Storlet-Dependency-Version': '1',
                        'X-Object-Meta-Storlet-Dependency-Permissions': '0755'}
            resp = self.get_request_response(target, 'POST', headers=sheaders)
            self.assertEqual('202 Accepted', resp.status)

    def test_GET_dependency(self):
        target = '/v1/AUTH_a/dependency/dependency'
        sheaders = {'X-Object-Meta-Storlet-Dependency-Version': '1',
                    'X-Object-Meta-Storlet-Dependency-Permissions': '0755'}
        self.base_app.register('GET', target, HTTPOk, headers=sheaders,
                               body=b'FAKE APP')

        with storlet_enabled():
            resp = self.get_request_response(target, 'GET')
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE APP', resp.body)
            for key in sheaders:
                self.assertEqual(sheaders[key], resp.headers.get(key))

    def test_storlets_with_invalid_method(self):
        with storlet_enabled():
            headers = {'X-Run-Storlet': 'Storlet-1.0.jar'}
            resp = self.get_request_response('/v1/AUTH_a/c/o',
                                             '_parse_vaco',
                                             headers=headers)
            self.assertEqual('405 Method Not Allowed', resp.status)

    def test_POST_storlet_acl_bad_request(self):
        # Test wrong target elements
        headers = {'X-Storlet-Container-Read': 'aa',
                   'X-Storlet-Name': 'bb'}
        target = '/v1/AUTH_a/c/o'
        self.base_app.register('POST', target, HTTPNoContent, body=b'')
        with storlet_enabled():
            resp = self.get_request_response(target, 'POST', headers=headers)
            self.assertEqual('204 No Content', resp.status)

        # Test wrong target containers
        targets = ['/v1/AUTH_a/dependency/', '/v1/AUTH_a/storlet/']
        with storlet_enabled():
            for target in targets:
                resp = self.get_request_response(target, 'POST',
                                                 headers=headers)
                self.assertEqual('400 Bad Request', resp.status)
                msg = b'storlet ACL update cannot be a storlet container'
                self.assertEqual(msg, resp.body)

        # Test wrong headers content
        hlist = [{'X-Storlet-Container-Read': ''},
                 {'X-Storlet-Container-Read': 'a'},
                 {'X-Storlet-Container-Read': 'a',
                  'X-Storlet-Name': ''}]
        target = '/v1/AUTH_a/c/'
        with storlet_enabled():
            for headers in hlist:
                resp = self.get_request_response(target, 'POST',
                                                 headers=headers)
                self.assertEqual('400 Bad Request', resp.status)
                self.assertIn(b'missing a mandatory header', resp.body)

        # Test wrong ACL content
        hlist = [{'X-Storlet-Container-Read': 'a,b',
                  'X-Storlet-Name': 'c'},
                 {'X-Storlet-Container-Read': 'a',
                  'X-Storlet-Name': 'b,c'}]
        target = '/v1/AUTH_a/c/'
        with storlet_enabled():
            for headers in hlist:
                resp = self.get_request_response(target, 'POST',
                                                 headers=headers)
                self.assertEqual('400 Bad Request', resp.status)
                self.assertIn(b'mulformed storlet or user name', resp.body)

    def test_POST_storlet_acl(self):
        target = '/v1/AUTH_a/c'
        head_headers = {'x-container-read': 'adam'}
        self.base_app.register('HEAD', target, HTTPNoContent, head_headers)
        self.base_app.register('POST', target, HTTPNoContent)

        headers = {'X-Storlet-Container-Read': 'a',
                   'X-Storlet-Name': 'b'}
        expected_read_acl = 'adam,.r:storlets.a_b'
        with storlet_enabled():
            resp = self.get_request_response(target, 'POST', headers=headers)
            self.assertEqual('204 No Content', resp.status)
            head_calls = self.base_app.get_calls('HEAD', target)
            self.assertEqual(1, len(head_calls))
            post_calls = self.base_app.get_calls('POST', target)
            self.assertEqual(1, len(post_calls))
            self.assertEqual(expected_read_acl,
                             post_calls[-1][2]['X-Container-Read'])

    def test_GET_with_invalid_referrer(self):
        headers = [{'Referer': REFERER_PREFIX},
                   {'Referer': 'a.b.%s' % REFERER_PREFIX},
                   {'Referer': 'a.b.%s.c.d' % REFERER_PREFIX},
                   {'Referer': '//%s' % REFERER_PREFIX},
                   {'Referer': '%s.c.d' % REFERER_PREFIX}]
        targets = ['/v1/AUTH_a/c', '/v1/AUTH_a/c/o']
        ops = ['GET', 'HEAD']
        for target, op, header in itertools.product(targets, ops, headers):
            resp = self.get_request_response(target, op, headers=header)
            self.assertEqual('403 Forbidden', resp.status)


class TestStorletProxyHandler(unittest.TestCase):
    def setUp(self):
        self.handler_class = StorletProxyHandler
        self.conf = create_handler_config('proxy')
        self.gateway_conf = {}

    def test_init_handler(self):
        req = Request.blank(
            '/v1/acc/cont/obj', environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storlet-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        with storlet_enabled():
            handler = self.handler_class(
                req, self.conf, self.gateway_conf,
                mock.MagicMock(), mock.MagicMock())

        self.assertEqual('/v1/acc/cont/obj', handler.request.path)
        self.assertEqual('v1', handler.api_version)
        self.assertEqual('acc', handler.account)
        self.assertEqual('cont', handler.container)
        self.assertEqual('obj', handler.obj)

        # overwrite the request
        # TODO(kota_): is it good to raise an error immediately?
        #              or re-validate the req again?
        req = Request.blank(
            '/v2/acc2/cont2/obj2', environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storlet-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        handler.request = req
        self.assertEqual('/v2/acc2/cont2/obj2', handler.request.path)
        self.assertEqual('v2', handler.api_version)
        self.assertEqual('acc2', handler.account)
        self.assertEqual('cont2', handler.container)
        self.assertEqual('obj2', handler.obj)

        # no direct assignment allowed
        with self.assertRaises(AttributeError):
            handler.api_version = '1'

        with self.assertRaises(AttributeError):
            handler.account = 'acc'

        with self.assertRaises(AttributeError):
            handler.container = 'cont'

        with self.assertRaises(AttributeError):
            handler.obj = 'obj'

    def test_remove_storlet_headers(self):
        req = Request.blank(
            '/v1/acc/cont/obj', environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storlet-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        with storlet_enabled():
            handler = self.handler_class(
                req, self.conf, self.gateway_conf,
                mock.MagicMock(), mock.MagicMock())

        headers = {'X-Storlet-Key1': 'Value1',
                   'X-Key2': 'Value2',
                   'X-Object-Meta-Storlet-Key3': 'Value3',
                   'X-Object-Meta-Key4': 'Value4'}
        handler._remove_storlet_headers(headers)

        self.assertNotIn('X-Storlet-Key1', headers)
        self.assertEqual('Value2', headers['X-Key2'])
        self.assertNotIn('X-Object-Meta-Storlet-Key3', headers)
        self.assertEqual('Value4', headers['X-Object-Meta-Key4'])

    def test_get_storlet_invocation_options(self):
        req = Request.blank(
            '/v1/acc/cont/obj',
            environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Run-Storlet': 'Storlet-1.0.jar',
                     'X-Storlet-Foo': 'baa'})
        with storlet_enabled():
            handler = self.handler_class(
                req, self.conf, self.gateway_conf, mock.MagicMock(),
                mock.MagicMock())

        options = handler._get_storlet_invocation_options(req)
        self.assertEqual('baa', options['storlet_foo'])
        self.assertFalse(options['generate_log'])

        req = Request.blank(
            '/v1/acc/cont/obj',
            environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Run-Storlet': 'Storlet-1.0.jar',
                     'X-Storlet-Foo': 'baa',
                     'X-Storlet-Generate-Log': 'True'})
        with storlet_enabled():
            handler = self.handler_class(
                req, self.conf, self.gateway_conf, mock.MagicMock(),
                mock.MagicMock())

        options = handler._get_storlet_invocation_options(req)
        self.assertEqual('baa', options['storlet_foo'])
        self.assertTrue(options['generate_log'])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import unittest
from unittest import mock

from storlets.sbus import command as sbus_cmd
from storlets.sbus.file_description import SBUS_FD_SERVICE_OUT
from storlets.sbus.datagram import SBusFileDescriptor, SBusServiceDatagram
from storlets.agent.common.server import EXIT_SUCCESS, command_handler, \
    CommandResponse, CommandFailure, CommandSuccess, SBusServer
from tests.unit import FakeLogger


class TestCommandResponse(unittest.TestCase):
    def test_init(self):
        resp = CommandResponse(True, 'ok')
        self.assertTrue(resp.status)
        self.assertEqual('ok', resp.message)
        self.assertTrue(resp.iterable)
        self.assertIsNone(resp.task_id)

        resp = CommandResponse(False, 'error', False)
        self.assertFalse(resp.status)
        self.assertEqual('error', resp.message)
        self.assertFalse(resp.iterable)
        self.assertIsNone(resp.task_id)

        resp = CommandResponse(True, 'ok', task_id='foo')
        self.assertTrue(resp.status)
        self.assertEqual('ok', resp.message)
        self.assertTrue(resp.iterable)
        self.assertEqual('foo', resp.task_id)

    def test_report_message(self):
        resp = CommandResponse(True, 'msg', True)
        self.assertEqual({'status': True, 'message': 'msg'},
                         json.loads(resp.report_message))

        resp = CommandResponse(True, 'msg', True, 'foo')
        self.assertEqual({'status': True, 'message': 'msg', 'task_id': 'foo'},
                         json.loads(resp.report_message))


class TestCommandSuccess(unittest.TestCase):
    def test_init(self):
        resp = CommandSuccess('ok')
        self.assertTrue(resp.status)
        self.assertEqual('ok', resp.message)
        self.assertTrue(resp.iterable)


class TestCommandFailure(unittest.TestCase):
    def test_init(self):
        resp = CommandFailure('error')
        self.assertFalse(resp.status)
        self.assertEqual('error', resp.message)
        self.assertTrue(resp.iterable)


class TestCommandHandler(unittest.TestCase):
    def test_command_handler(self):

        @command_handler
        def test_func():
            pass

        self.assertTrue(hasattr(test_func, 'is_command_handler'))
        self.assertTrue(test_func.is_command_handler)


class TestSBusServer(unittest.TestCase):
    def setUp(self):
        self.logger = FakeLogger()
        self.sbus_path = 'path/to/pipe'
        self.server = SBusServer(self.sbus_path, self.logger)

    def test_get_handler(self):
        # halt
        self.assertEqual(
            self.server.halt, self.server.get_handler(sbus_cmd.SBUS_CMD_HALT))
        # ping
        self.assertEqual(
            self.server.ping, self.server.get_handler(sbus_cmd.SBUS_CMD_PING))

        # daemon status is defined as sbus command, but the command is not
        # implemented in SBusServer
        with self.assertRaises(ValueError):
            self.server.get_handler('SBUS_CMD_DAEMON_STATUS')

        # invalid command
        with self.assertRaises(ValueError):
            self.server.get_handler('FOO')
        # unknown command
        with self.assertRaises(ValueError):
            self.server.get_handler('SBUS_CMD_UNKNOWN')


def create_fake_sbus_class(scenario):
    """
    :param scenario: a list of tuples [(command, return_value)]
    """
    class FakeSBus(object):
        """
        Fake SBus communication according fed scenario.
        The scenario must be ordered by calls
        """
        def __init__(self):
            self.scenario = scenario
            self.called = []

        def _get_fake_response(self, command, *args):
            self.called.append((command, args))
            try:
                ret_val = self.scenario.pop(0)
            except IndexError:
                raise AssertionError(
                    'Daemon calls SBus command more than expected,'
                    ' need "%s" call at least' % command)

            if ret_val[0] != command:
                raise AssertionError(
                    "Expected scenario doesn't happen, actual daemon calls %s"
                    "for SBus but %d is expected" % (command, ret_val[0]))
            return ret_val[1]

        def listen(self, fd, timeout=0):
            return self._get_fake_response('listen', fd)

        def command(self):
            return self._get_fake_response('command')

        def create(self, sbus_path):
            return self._get_fake_response('create', sbus_path)

        def receive(self, fd):
            return self._get_fake_response('receive', fd)

    return FakeSBus


class TestSBusServerMain(unittest.TestCase):

    def _get_test_server(self):
        return SBusServer(self.sbus_path, self.logger)

    def setUp(self):
        self.logger = FakeLogger()
        self.sbus_path = 'fake_path'
        self.server = self._get_test_server()

    def _test_main_loop_stop(self, stop_command):
        sfds = [SBusFileDescriptor(SBUS_FD_SERVICE_OUT, 1)]
        scenario = [
            ('create', 1),
            ('listen', 1),
            ('receive', SBusServiceDatagram(command=stop_command, sfds=sfds,
                                            params=None, task_id=None)),
        ]

        fake_sbus_class = create_fake_sbus_class(scenario)
        with mock.patch('storlets.agent.common.server.SBus', fake_sbus_class):
            with mock.patch('os.fdopen'):
                ret = self.server.main_loop()

        self.assertEqual(EXIT_SUCCESS, ret)
        # sanity for no error and no warning
        self.assertEqual([], self.logger.get_log_lines('error'))
        self.assertEqual([], self.logger.get_log_lines('warn'))

    def _test_main_loop_timeout(self):
        scenario = [
            ('create', 1),
            ('listen', 0),
            ('listen', 0),
            ('listen', 0),
            ('listen', 0)
        ]

        self.server.listen_timeout = self.server.loop_timeout * 4

        fake_sbus_class = create_fake_sbus_class(scenario)
        with mock.patch('storlets.agent.common.server.SBus', fake_sbus_class):
            with mock.patch('os.fdopen'):
                ret = self.server.main_loop()

        self.assertEqual(EXIT_SUCCESS, ret)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

SBUS_CMD_PREFIX = 'SBUS_CMD_'

SBUS_CMD_HALT = 'SBUS_CMD_HALT'
SBUS_CMD_EXECUTE = 'SBUS_CMD_EXECUTE'
SBUS_CMD_START_DAEMON = 'SBUS_CMD_START_DAEMON'
SBUS_CMD_STOP_DAEMON = 'SBUS_CMD_STOP_DAEMON'
SBUS_CMD_DAEMON_STATUS = 'SBUS_CMD_DAEMON_STATUS'
SBUS_CMD_STOP_DAEMONS = 'SBUS_CMD_STOP_DAEMONS'
SBUS_CMD_PING = 'SBUS_CMD_PING'
SBUS_CMD_CANCEL = 'SBUS_CMD_CANCEL'
SBUS_CMD_NOP = 'SBUS_CMD_NOP'



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

SBUS_FD_INPUT_OBJECT = 'SBUS_FD_INPUT_OBJECT'
SBUS_FD_OUTPUT_OBJECT = 'SBUS_FD_OUTPUT_OBJECT'
SBUS_FD_OUTPUT_OBJECT_METADATA = 'SBUS_FD_OUTPUT_OBJECT_METADATA'
SBUS_FD_OUTPUT_OBJECT_AND_METADATA = 'SBUS_FD_OUTPUT_OBJECT_AND_METADATA'
SBUS_FD_LOGGER = 'SBUS_FD_LOGGER'
SBUS_FD_OUTPUT_CONTAINER = 'SBUS_FD_OUTPUT_CONTAINER'
SBUS_FD_SERVICE_OUT = 'SBUS_FD_SERVICE_OUT'



# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
import errno
import json
import os
import unittest
from unittest import mock

from storlets.sbus.client.exceptions import SBusClientSendError, \
    SBusClientMalformedResponse
from storlets.sbus.client import SBusClient


@contextmanager
def _mock_sbus(send_status=0):
    with mock.patch('storlets.sbus.client.client.SBus.send') as fake_send:
        fake_send.return_value = send_status
        yield


@contextmanager
def _mock_os_pipe(bufs):
    class FakeFd(object):
        def __init__(self, rbuf=''):
            self.rbuf = rbuf
            self.closed = False

        def read(self, size):
            size = min(len(self.rbuf), size)
            ret = self.rbuf[:size]
            self.rbuf = self.rbuf[size:]
            return ret

        def close(self):
            if self.closed:
                raise OSError(errno.EBADF, os.strerror(errno.EBADF))
            self.closed = True

    def fake_os_read(fd, size):
        return fd.read(size)

    def fake_os_close(fd):
        fd.close()

    pipes = [(FakeFd(buf), FakeFd()) for buf in bufs]
    pipe_generator = iter(pipes)

    def mock_os_pipe():
        try:
            return next(pipe_generator)
        except StopIteration:
            raise AssertionError('pipe called more than expected')

    with mock.patch('storlets.sbus.client.client.os.pipe', mock_os_pipe), \
            mock.patch('storlets.sbus.client.client.os.read', fake_os_read), \
            mock.patch('storlets.sbus.client.client.os.close', fake_os_close):
        yield pipes


class TestSBusClient(unittest.TestCase):
    def setUp(self):
        self.pipe_path = 'pipe_path'
        self.client = SBusClient(self.pipe_path, 4)

    def test_parse_response(self):
        raw_resp = json.dumps({'status': True, 'message': 'OK'})
        resp = self.client._parse_response(raw_resp)
        self.assertTrue(resp.status)
        self.assertEqual('OK', resp.message)
        self.assertIsNone(resp.task_id)

        raw_resp = json.dumps({'status': True, 'message': 'OK',
                               'task_id': 'SOMEID'})
        resp = self.client._parse_response(raw_resp)
        self.assertTrue(resp.status)
        self.assertEqual('OK', resp.message)
        self.assertEqual('SOMEID', resp.task_id)

        raw_resp = json.dumps({'status': False, 'message': 'ERROR'})
        resp = self.client._parse_response(raw_resp)
        self.assertFalse(resp.status)
        self.assertEqual('ERROR', resp.message)
        self.assertIsNone(resp.task_id)

        raw_resp = json.dumps({'status': True, 'message': 'Sample:Message'})
        resp = self.client._parse_response(raw_resp)
        self.assertTrue(resp.status)
        self.assertEqual('Sample:Message', resp.message)
        self.assertIsNone(resp.task_id)

        with self.assertRaises(SBusClientMalformedResponse):
            self.client._parse_response('Foo')

        raw_resp = json.dumps({'status': True})
        with self.assertRaises(SBusClientMalformedResponse):
            self.client._parse_response(raw_resp)

        raw_resp = json.dumps({'message': 'foo'})
        with self.assertRaises(SBusClientMalformedResponse):
            self.client._parse_response(raw_resp)

    def _check_all_pipes_closed(self, pipes):
        # Make sure that pipes are not empty
        self.assertGreater(len(pipes), 0)

        for _pipe in pipes:
            self.assertTrue(_pipe[0].closed)
            self.assertTrue(_pipe[1].closed)

    def _test_service_request(self, method, *args, **kwargs):
        raw_resp = json.dumps(
            {'status': True, 'message': 'OK'}).encode("utf-8")
        with _mock_os_pipe([raw_resp]) as pipes, _mock_sbus(0):
            resp = method(*args, **kwargs)
            self.assertTrue(resp.status)
            self.assertEqual('OK', resp.message)
            self._check_all_pipes_closed(pipes)

        raw_resp = json.dumps(
            {'status': False, 'message': 'ERROR'}).encode("utf-8")
        with _mock_os_pipe([raw_resp]) as pipes, _mock_sbus(0):
            resp = method(*args, **kwargs)
            self.assertFalse(resp.status)
            self.assertEqual('ERROR', resp.message)
            self._check_all_pipes_closed(pipes)

        raw_resp = json.dumps(
            {'status': True, 'message': 'OK'}).encode("utf-8")
        with _mock_os_pipe([raw_resp]) as pipes, _mock_sbus(-1):
            with self.assertRaises(SBusClientSendError):
                method(*args, **kwargs)
            self._check_all_pipes_closed(pipes)

        # TODO(takashi): Add IOError case

        with _mock_os_pipe([b'Foo']) as pipes, _mock_sbus(0):
            with self.assertRaises(SBusClientMalformedResponse):
                method(*args, **kwargs)
            self._check_all_pipes_closed(pipes)

    def test_ping(self):
        self._test_service_request(self.client.ping)

    def test_start_daemon(self):
        self._test_service_request(
            self.client.start_daemon, 'java', 'path/to/storlet',
            'storleta', 'path/to/uds', 'INFO', '10', '11')

    def test_stop_daemon(self):
        self._test_service_request(self.client.stop_daemon, 'storleta')

    def test_stop_daemons(self):
        self._test_service_request(self.client.stop_daemons)

    def test_halt(self):
        self._test_service_request(self.client.halt)

    def test_daemon_status(self):
        self._test_service_request(self.client.daemon_status, 'storleta')

    def test_cancel(self):
        self._test_service_request(self.client.cancel, 'taskid')


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

import docker.client
import docker.errors
import docker.models.containers

from storlets.sbus.client import SBusResponse
from storlets.sbus.client.exceptions import SBusClientIOError, \
    SBusClientMalformedResponse, SBusClientSendError
from storlets.gateway.common.exceptions import StorletRuntimeException, \
    StorletTimeout
from storlets.gateway.gateways.docker.runtime import DockerRunTimeSandbox
from tests.unit import FakeLogger


class TestDockerRunTimeSandbox(unittest.TestCase):
    def setUp(self):
        self.logger = FakeLogger()
        # TODO(takashi): take these values from config file
        self.conf = {'container_image_namespace': 'localhost:5001',
                     'default_container_image_name': 'defaultimage'}
        self.scope = '0123456789abc'
        self.sbox = DockerRunTimeSandbox(self.scope, self.conf, self.logger)

    def test_ping(self):
        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.ping') as ping:
            ping.return_value = SBusResponse(True, 'OK')
            self.assertTrue(self.sbox.ping())

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.ping') as ping:
            ping.return_value = SBusResponse(False, 'Error')
            self.assertFalse(self.sbox.ping())

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.ping') as ping:
            ping.side_effect = SBusClientSendError()
            self.assertFalse(self.sbox.ping())

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.ping') as ping:
            ping.side_effect = SBusClientMalformedResponse()
            self.assertFalse(self.sbox.ping())

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.ping') as ping:
            ping.side_effect = SBusClientIOError()
            self.assertFalse(self.sbox.ping())

    def test_wait(self):
        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.ping') as ping, \
                mock.patch('storlets.gateway.gateways.container.runtime.'
                           'time.sleep') as sleep:
            ping.return_value = SBusResponse(True, 'OK')
            self.sbox.wait()
            self.assertEqual(sleep.call_count, 0)

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'SBusClient.ping') as ping, \
            mock.patch('storlets.gateway.gateways.container.runtime.'
                       'time.sleep') as sleep:
            ping.side_effect = [SBusResponse(False, 'Error'),
                                SBusResponse(True, 'OK')]
            self.sbox.wait()
            self.assertEqual(sleep.call_count, 1)

        # TODO(takashi): should test timeout case

    def test__restart(self):
        # storlet container is not running
        with mock.patch('storlets.gateway.gateways.docker.runtime.'
                        'docker.from_env') as docker_from_env:
            mock_client = mock.MagicMock(spec_set=docker.client.DockerClient)
            mock_containers = mock.MagicMock(
                spec_set=docker.models.containers.ContainerCollection)
            mock_client.containers = mock_containers
            mock_containers.get.side_effect = \
                docker.errors.NotFound('container is not found')
            docker_from_env.return_value = mock_client

            self.sbox._restart('storlet_image')
            self.assertEqual(1, mock_containers.get.call_count)
            self.assertEqual(0, mock_containers.list.call_count)
            self.assertEqual(1, mock_containers.run.call_count)

        # storlet container is running
        with mock.patch('storlets.gateway.gateways.docker.runtime.'
                        'docker.from_env') as docker_from_env:
            mock_client = mock.MagicMock(spec_set=docker.client.DockerClient)
            mock_containers = mock.MagicMock(
                spec_set=docker.models.containers.ContainerCollection)
            mock_client.containers = mock_containers
            mock_container = \
                mock.MagicMock(spec_set=docker.models.containers.Container)
            mock_containers.get.return_value = mock_container
            docker_from_env.return_value = mock_client

            self.sbox._restart('storlet_image')
            self.assertEqual(1, mock_containers.get.call_count)
            self.assertEqual(1, mock_container.stop.call_count)
            self.assertEqual(0, mock_containers.list.call_count)
            self.assertEqual(1, mock_containers.run.call_count)

        # get failed
        with mock.patch('storlets.gateway.gateways.docker.runtime.'
                        'docker.from_env') as docker_from_env:
            mock_client = mock.MagicMock(spec_set=docker.client.DockerClient)
            mock_containers = mock.MagicMock(
                spec_set=docker.models.containers.ContainerCollection)
            mock_client.containers = mock_containers
            mock_containers.get.side_effect = \
                docker.errors.APIError('api error')
            docker_from_env.return_value = mock_client

            with self.assertRaises(StorletRuntimeException):
                self.sbox._restart('storlet_image')
            self.assertEqual(1, mock_containers.get.call_count)
            self.assertEqual(0, mock_containers.run.call_count)

        # stop failed
        with mock.patch('storlets.gateway.gateways.docker.runtime.'
                        'docker.from_env') as docker_from_env:
            mock_client = mock.MagicMock(spec_set=docker.client.DockerClient)
            mock_containers = mock.MagicMock(
                spec_set=docker.models.containers.ContainerCollection)
            mock_client.containers = mock_containers
            mock_container = \
                mock.MagicMock(spec_set=docker.models.containers.Container)
            mock_containers.get.return_value = mock_container
            mock_container.stop.side_effect = \
                docker.errors.APIError('api error')
            docker_from_env.return_value = mock_client

            with self.assertRaises(StorletRuntimeException):
                self.sbox._restart('storlet_image')
            self.assertEqual(1, mock_containers.get.call_count)
            self.assertEqual(1, mock_container.stop.call_count)
            self.assertEqual(0, mock_containers.list.call_count)
            self.assertEqual(0, mock_containers.run.call_count)

        # run failed
        with mock.patch('storlets.gateway.gateways.docker.runtime.'
                        'docker.from_env') as docker_from_env:
            mock_client = mock.MagicMock(spec_set=docker.client.DockerClient)
            mock_containers = mock.MagicMock(
                spec_set=docker.models.containers.ContainerCollection)
            mock_containers.run.side_effect = \
                docker.errors.APIError('api error')
            mock_client.containers = mock_containers
            mock_container = \
                mock.MagicMock(spec_set=docker.models.containers.Container)
            mock_containers.get.return_value = mock_container
            docker_from_env.return_value = mock_client

            with self.assertRaises(StorletRuntimeException):
                self.sbox._restart('storlet_image')
            self.assertEqual(1, mock_containers.get.call_count)
            self.assertEqual(1, mock_container.stop.call_count)
            self.assertEqual(0, mock_containers.list.call_count)
            self.assertEqual(1, mock_containers.run.call_count)

        # Set the limit
        self.sbox.max_containers_per_node = 2

        with mock.patch('storlets.gateway.gateways.docker.runtime.'
                        'docker.from_env') as docker_from_env:
            mock_client = mock.MagicMock(spec_set=docker.client.DockerClient)
            mock_containers = mock.MagicMock(
                spec_set=docker.models.containers.ContainerCollection)
            mock_client.containers = mock_containers
            mock_container = \
                mock.MagicMock(spec_set=docker.models.containers.Container)
            mock_containers.get.return_value = mock_container
            mock_containers.list.return_value = [mock.MagicMock()]
            docker_from_env.return_value = mock_client

            self.sbox._restart('storlet_image')
            self.assertEqual(1, mock_containers.get.call_count)
            self.assertEqual(1, mock_container.stop.call_count)
            self.assertEqual(1, mock_containers.list.call_count)
            self.assertEqual(1, mock_containers.run.call_count)

        with mock.patch('storlets.gateway.gateways.docker.runtime.'
                        'docker.from_env') as docker_from_env:
            mock_client = mock.MagicMock(spec_set=docker.client.DockerClient)
            mock_containers = mock.MagicMock(
                spec_set=docker.models.containers.ContainerCollection)
            mock_client.containers = mock_containers
            mock_container = \
                mock.MagicMock(spec_set=docker.models.containers.Container)
            mock_containers.get.return_value = mock_container
            mock_containers.list.return_value = [mock.MagicMock()] * 2
            docker_from_env.return_value = mock_client

            with self.assertRaises(StorletRuntimeException):
                self.sbox._restart('storlet_image')
            self.assertEqual(1, mock_containers.get.call_count)
            self.assertEqual(1, mock_container.stop.call_count)
            self.assertEqual(1, mock_containers.list.call_count)
            self.assertEqual(0, mock_containers.run.call_count)

    def test_restart(self):
        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'RunTimePaths.create_host_pipe_dir') as pipe_dir, \
                mock.patch('storlets.gateway.gateways.docker.runtime.'
                           'DockerRunTimeSandbox._restart') as _restart, \
                mock.patch('storlets.gateway.gateways.docker.runtime.'
                           'DockerRunTimeSandbox.wait') as wait:
            self.sbox.restart()
            self.assertEqual(1, pipe_dir.call_count)
            self.assertEqual(1, _restart.call_count)
            self.assertEqual((self.scope,), _restart.call_args.args)
            self.assertEqual(1, wait.call_count)

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'RunTimePaths.create_host_pipe_dir') as pipe_dir, \
                mock.patch('storlets.gateway.gateways.docker.runtime.'
                           'DockerRunTimeSandbox._restart') as _restart, \
                mock.patch('storlets.gateway.gateways.docker.runtime.'
                           'DockerRunTimeSandbox.wait') as wait:
            _restart.side_effect = [StorletRuntimeException(), None]
            self.sbox.restart()
            self.assertEqual(1, pipe_dir.call_count)
            self.assertEqual(2, _restart.call_count)
            self.assertEqual((self.scope,),
                             _restart.call_args_list[0].args)
            self.assertEqual(('defaultimage',),
                             _restart.call_args_list[1].args)
            self.assertEqual(1, wait.call_count)

        with mock.patch('storlets.gateway.gateways.container.runtime.'
                        'RunTimePaths.create_host_pipe_dir') as pipe_dir, \
                mock.patch('storlets.gateway.gateways.docker.runtime.'
                           'DockerRunTimeSandbox._restart') as _restart, \
                mock.patch('storlets.gateway.gateways.docker.runtime.'
                           'DockerRunTimeSandbox.wait') as wait:
            _restart.side_effect = StorletTimeout()
            with self.assertRaises(StorletRuntimeException):
                self.sbox.restart()
            self.assertEqual(1, pipe_dir.call_count)
            self.assertEqual(1, _restart.call_count)
            self.assertEqual((self.scope,), _restart.call_args.args)
            self.assertEqual(0, wait.call_count)

    def test_get_storlet_classpath(self):
        storlet_id = 'Storlet.jar'
        storlet_main = 'org.openstack.storlet.Storlet'
        dependencies = ['dep1', 'dep2']
        self.assertEqual(
            '/home/swift/org.openstack.storlet.Storlet/Storlet.jar:'
            '/home/swift/org.openstack.storlet.Storlet/dep1:'
            '/home/swift/org.openstack.storlet.Storlet/dep2',
            self.sbox._get_storlet_classpath(storlet_main, storlet_id,
                                             dependencies),)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SimpleStorlet(object):
    def __init__(self, logger):
        self.logger = logger

    def __call__(self, in_files, out_files, params):
        """
        The function called for storlet invocation

        :param in_files: a list of StorletInputFile
        :param out_files: a list of StorletOutputFile
        :param params: a dict of request parameters
        """
        self.logger.debug('Returning metadata')
        metadata = in_files[0].get_metadata()
        metadata['test'] = 'simple'
        out_files[0].set_metadata(metadata)

        self.logger.debug('Start to return object data')
        while True:
            buf = in_files[0].read(16)
            if not buf:
                break
            self.logger.debug('Received %d bytes' % len(buf))
            self.logger.debug('Writing back %d bytes' % len(buf))
            out_files[0].write(buf)
        self.logger.debug('Complete')
        in_files[0].close()
        out_files[0].close()



# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import unittest
from io import StringIO

from storlets.agent.common.utils import get_logger


class TestLogger(unittest.TestCase):

    def test_get_logger(self):
        sio = StringIO()
        logger = logging.getLogger('CONT #abcdef: test')
        logger.addHandler(logging.StreamHandler(sio))

        # set log level as INFO
        logger = get_logger('test', 'INFO', 'abcdef')
        self.assertEqual(logging.INFO, logger.level)
        # INFO message is recorded with INFO log level
        logger.info('test1')
        self.assertEqual('test1\n', sio.getvalue())
        # DEBUG message is not recorded with INFO log level
        logger.debug('test2')
        self.assertEqual('test1\n', sio.getvalue())

        # set log level as DEBUG
        logger = get_logger('test', 'DEBUG', 'abcdef')
        self.assertEqual(logging.DEBUG, logger.level)
        # DEBUG message is recorded with DEBUG log level
        logger.debug('test3')
        self.assertEqual('test1\ntest3\n', sio.getvalue())

        # If the level parameter is unknown, use ERROR as log level
        logger = get_logger('test', 'foo', 'abcdef')
        self.assertEqual(logging.ERROR, logger.level)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
import unittest
from storlets.gateway.common.exceptions import StorletLoggerError
from storlets.gateway.common.logger import StorletLogger


class TestStorletLogger(unittest.TestCase):

    def setUp(self):
        self.log_dir = tempfile.mkdtemp()
        self.log_path = tempfile.mktemp(dir=self.log_dir)
        self.logger = StorletLogger(self.log_path)

    def tearDown(self):
        if os.path.isdir(self.log_dir):
            shutil.rmtree(self.log_dir)

    def test_open_close(self):
        self.assertIsNone(self.logger._file)
        self.logger.open()
        self.assertIsNotNone(self.logger._file)
        self.assertTrue(os.path.isfile(self.log_path))
        self.logger.close()
        self.assertIsNone(self.logger._file)

        # Make sure that log_dir is deleted
        shutil.rmtree(self.log_dir)

        # As log_dir does not exists, open should create it
        self.logger.open()
        self.assertIsNotNone(self.logger._file)
        self.assertTrue(os.path.isdir(self.log_dir))
        self.assertTrue(os.path.isfile(self.log_path))
        self.logger.close()
        self.assertIsNone(self.logger._file)

        # opened twice
        self.logger.open()
        with self.assertRaises(StorletLoggerError):
            self.logger.open()
        self.logger.close()

        # closed twice
        self.logger.open()
        self.logger.close()
        with self.assertRaises(StorletLoggerError):
            self.logger.close()

    def test_getfd(self):
        with self.assertRaises(StorletLoggerError):
            self.logger.getfd()

        self.logger.open()
        self.assertIsNotNone(self.logger.getfd())
        self.logger.close()

    def test_getsize(self):
        self.logger.open()
        self.logger._file.write('a' * 1024)
        self.logger.close()
        self.assertEqual(1024, self.logger.getsize())
